searchData={"content_type":"text/markdown","items":[{"doc":"Endo is a library which allows you to reflect on the tables, columns, indexes, and\nother data stored within a given Ecto Repo.","ref":"Endo.html","title":"Endo","type":"module"},{"doc":"To get started with Endo, all you need is a supported Ecto Repo from your application,\nas well as the Endo dependency.\n\nCurrently, Endo only supports the `Ecto.Adapters.Postgres` adapter, but Endo was built\nto be adapter agnostic, so additional adapters may be implemented in the future.\n\nGiven an Ecto Repo whose adapter is `Ecto.Adapters.Postgres`, you can execute Endo with\nthe following snippet:\n\n```elixir\nEndo.list_tables(MyApp.Repo)\n```\n\nThis default invokation will return a list of all public tables in your repo's database,\nas well as metadata pertaining to said tables.\n\nAt the time of writing, table metadata includes:\n\n- The name of the given table.\n- A list of columns defined in the given table.\n- A list of indexes defined in the given table.\n- A list of associations, if any, sourced from the given table.\n- The repo adapter, for reference.\n\nIn the future, we plan on implementing a list of constraints, as well as surfacing additional\nstats and metrics which we can lift from the underlying adapters.\n\n> ##","ref":"Endo.html#module-basic-usage","title":"Basic Usage - Endo","type":"module"},{"doc":">\n> By default, Endo will return *all tables in the public schema* for a Postgres repo.\n> This will include a default `schema_migrations` table which is seeded by Ecto itself.\n>\n> Additionally, if you're using dependencies such as `Oban`, any tables created for these\n> dependencies may also be returned.","ref":"Endo.html#module-note-info","title":"Note {: .info} - Endo","type":"module"},{"doc":"By providing the optional second paramter to `Endo.list_tables/2`, you can instruct Endo to\nfetch tables matching an arbitrary set of criterion.\n\nOptions are ultimately passed down into the Endo adapter matching the given repo module, but\nin general the supported high-level options are:\n\n- `with_column`, instructs Endo to only return tables which define a column with the given name\n- `without_column`, instructs Endo to only return tables which do not define a column with the\n  given name\n- `with_foreign_key_constraint`, instructs Endo to only return tables which have a foreign key constraint\n  to the given table name\n- `without_foreign_key_constraint`, instructs Endo to only return tables which do not have a foreign\n  key constraint to the given table name\n- `with_index`, instructs Endo to only return tables with indexes on the given column or columns.\n  If a list of columns if given, then we look for a compound index that exactly matches the given ones\n  components and order.\n- `without_index`, instructs Endo to only return tables without indexes on the given column or columns.\n  Likewise, compound indexes must exactly match the given ones components and order.\n- `with_index_covering`, instructs Endo to only return tables with an index which covers the given\n  column(s). An index covering a given column(s) is defined as whether or not there exists a single index\n  or composite index which contains the given column(s) regardless of ordering.\n- `without_index_covering`, instructs Endo to only return tables without an index which covers the given\n  column(s). The same caveats as the above option apply.\n\nAdditionally, adapters are free to implement custom filters. The `Endo.Adapters.Postgres` adapter\nforwards any filters not matching the above list directly to direct SQL queries against the base\ninformation schema tables the adapter sources from.\n\nThis means that is it possible to perform direct queries such as `Endo.list_tables(MyApp.Repo, table_name: \"payments\")`.\nSee `Endo.Adapters.Postgres` and `Endo.Adapters.Postgres.Table` for a list of supported fields -- though as these\nare internal to Endo, do not count on these being stable across version upgrades.\n\nEndo also supports multiple filters being given in a single invokation. Multiple filters will apply\nin addition to one another, in the order which they are given. The same filter can be provided multiple\ntimes also.\n\nSee the following example for details:\n\n```elixir\nEndo.list_tables(MyApp.Repo, with_column: \"inserted_at\")\n# Returns list of tables defining an `inserted_at` column.\n\nEndo.list_tables(MyApp.Repo, with_column: \"inserted_at\", without_index: \"inserted_at\")\n# Returns list of tables defining an `inserted_at` column that does not index said column.\n\nEndo.list_tables(MyApp.Repo, with_column: \"inserted_at\", without_column: \"updated_at\")\n# Returns list of tables defining an `inserted_at` column while also not defining an `updated_at` column.\n\nEndo.list_tables(MyApp.Repo, with_foreign_key_constraint: \"cars\", without_index: \"car_id\")\n# Returns list of tables defining a relation to `cars`, without an index on said column\n\nEndo.list_tables(MyApp.Repo, with_index: [\"org_id\", \"location_id\", \"patient_id\"])\n# Returns list of tables defining a *composite index* on `(org_id, location_id, patient_id)`\n# Tables containing the same index in a different order, or a partial match will not be returned.\n\nEndo.list_tables(MyApp.Repo, with_index_covering: \"org_id\")\n# Returns list of tables defining any index which includes `org_id` (exact matches, or there exists\n# some composite index where `org_id` is a member).\n# Ordering does not matter.\n```\n\nAdditionally, many filters (such as `table_name`, `with_column`, etc.) can be given a regular expression which\nwill be attempted to be used by the Endo adapter of choice. This enables more complex queries to be performed,\nsuch as:\n\n```elixir\nEndo.list_tables(MyApp.Repo, with_column: ~r/(^|_)ID$/i)\n# Return all tables containing columns that either have a singular `id` column, or a `*_id` column.\n```","ref":"Endo.html#module-complex-lookups","title":"Complex Lookups - Endo","type":"module"},{"doc":"While Endo is designed to be able to adapt and connect to several different database backends, it is known that\ncertain features are very much database-engine specific.\n\nAs a result of this, each `Endo.Table.t()` also surfaces potentially adapter-specific metadata by way of its\n`metadata` key.\n\nA piece of `metadata` will be of type `Endo.Metadata.Postgres.t()` if it surfaces from the `Endo.Adapters.Postgres`\nadapter for example. Said metadata will thus be tailored to the underlying repo database engine.\n\nThis information is useful if one wants to build CI checks and other features on a lower level; i.e. writing\na check that enforces all tables in your application which don't have a primary key has `REPLICA IDENTITY FULL` set\nto enable Postgres to replicate to read replicas correctly.\n\nSee below for an example of such a check:\n\n```elixir\n[] =\n  Repo\n  |> Endo.list_tables()\n  |> Enum.reject(&Enum.any?(&1.indexes, fn index -> index.is_primary end))\n  |> Enum.reject(&(&1.metadata.replica_identity == \"FULL\"))\n```\n\nAs features get built out, we make no hard guarantees of keeping the `metadata` field stable, but efforts will\nof course be taken to mitigate unnecessary breaking changes from occuring.","ref":"Endo.html#module-adapter-specific-metadata","title":"Adapter-specific metadata - Endo","type":"module"},{"doc":"By default, Endo will assume that your application's tables are stored in the `public` schema. This is the\ndefault schema for Postgres, but it is possible to change this by setting the `:table_schema` config option\nin your application's config like so:\n\n```elixir\nconfig :endo, table_schema: \"custom_schema\"\n```\n\nIf set this way, Endo will use the given schema as the default schema for all queries. This can be overridden\non a per-query basis by passing the `:prefix` option to `Endo.list_tables/2` or `Endo.get_table/3` mimicing\nthe behavior of the `:prefix` option in `Ecto`:\n\n```elixir\n[%Endo.Table{}, ...] = Endo.list_tables(MyApp.Repo, prefix: \"custom_schema\")\n%Endo.Table{} = Endo.get_table(MyApp.Repo, \"special_table\", prefix: \"custom_schema\")\n```\n\nAdditionally, Endo will display the table schema of any given table via `%Endo.Table{schema: schema}`.\n\nFor runtime debugging, you can use the `Endo.table_schema/0` function to retrieve the current default schema.","ref":"Endo.html#module-setting-schema-prefixes-table-schemas","title":"Setting schema prefixes / table schemas - Endo","type":"module"},{"doc":"Given an Ecto Repo and a table name, tries to return an Endo Table or nil if said table does\nnot exist.\n\nInternally delegates to `list_tables/2` with the `table_name` option set.\nSee `list_tables/2` for more information.","ref":"Endo.html#get_table/3","title":"Endo.get_table/3","type":"function"},{"doc":"Given an Ecto Repo, returns a list of all tables, columns, associations, and indexes.\nTakes an optional keyword list of filters which filter said tables down to given constraints.","ref":"Endo.html#list_tables/2","title":"Endo.list_tables/2","type":"function"},{"doc":"Given Endo Column(s) or Endo Table(s), tries to populate a given `Endo.Column.t()`'s `Endo.Index.NotLoaded.t()`\nentries.\n\nSee `Endo.Index.load/2` for more information.","ref":"Endo.html#load_indexes/2","title":"Endo.load_indexes/2","type":"function"},{"doc":"Given a list of Endo Tables or a single Endo Table, tries to load the application-specific Ecto Schemas\nSee `Endo.Schema.load/1` for more information.","ref":"Endo.html#load_schemas/1","title":"Endo.load_schemas/1","type":"function"},{"doc":"Returns the default table schema used by Endo if not otherwise specified.","ref":"Endo.html#table_schema/0","title":"Endo.table_schema/0","type":"function"},{"doc":"Module defining the `Endo.Adapter` behaviour. Valid adapters will allow Endo to reflect\nupon a module implementing said adater","ref":"Endo.Adapter.html","title":"Endo.Adapter","type":"behaviour"},{"doc":"","ref":"Endo.Adapter.html#c:list_tables/2","title":"Endo.Adapter.list_tables/2","type":"callback"},{"doc":"","ref":"Endo.Adapter.html#c:to_endo/2","title":"Endo.Adapter.to_endo/2","type":"callback"},{"doc":"Adapter module implementing the ability for Endo to reflect upon any\nPostgres-based Ecto Repo.\n\nSee `Endo` documentation for list of features.\n\nIn future, parts of `Endo`'s top level documentation may be moved here, but as\nthis is the only supported adapter at the time of writing, this isn't the case.","ref":"Endo.Adapters.Postgres.html","title":"Endo.Adapters.Postgres","type":"module"},{"doc":"","ref":"Endo.Adapters.Postgres.html#list_tables/2","title":"Endo.Adapters.Postgres.list_tables/2","type":"function"},{"doc":"","ref":"Endo.Adapters.Postgres.html#to_endo/2","title":"Endo.Adapters.Postgres.to_endo/2","type":"function"},{"doc":"Utility module for taking a Postgres table and exposing adapter specific metadata to Endo","ref":"Endo.Adapters.Postgres.Metadata.html","title":"Endo.Adapters.Postgres.Metadata","type":"module"},{"doc":"","ref":"Endo.Adapters.Postgres.Metadata.html#derive!/1","title":"Endo.Adapters.Postgres.Metadata.derive!/1","type":"function"},{"doc":"Association metadata for a given table's associations","ref":"Endo.Association.html","title":"Endo.Association","type":"module"},{"doc":"","ref":"Endo.Association.html#t:t/0","title":"Endo.Association.t/0","type":"type"},{"doc":"Column metadata for a given table's columns","ref":"Endo.Column.html","title":"Endo.Column","type":"module"},{"doc":"","ref":"Endo.Column.html#t:t/0","title":"Endo.Column.t/0","type":"type"},{"doc":"","ref":"Endo.Column.Postgres.Type.Metadata.html","title":"Endo.Column.Postgres.Type.Metadata","type":"module"},{"doc":"","ref":"Endo.Column.Postgres.Type.Metadata.html#derive!/1","title":"Endo.Column.Postgres.Type.Metadata.derive!/1","type":"function"},{"doc":"","ref":"Endo.Column.Postgres.Type.Metadata.html#t:t/0","title":"Endo.Column.Postgres.Type.Metadata.t/0","type":"type"},{"doc":"Index metadata for a given table's indexes","ref":"Endo.Index.html","title":"Endo.Index","type":"module"},{"doc":"Tries to load a given `Endo.Column.t()`'s `indexes` field.\n\nCan take multiple inputs:\n  - A single `Endo.Table.t()`\n  - A list of `Endo.Table.t()`s\n  - A single `Endo.Column.t()`\n  - A list of `Endo.Column.t()`s\n\nPlease note that given `Endo.Column.t()` structs, additional `Endo` lookups are necessary. Thus, for the best\nperformance, it will be more optimal to pass in `Endo.Table.t()` structs if possible.\n\nWill raise an error if given a mixed list of `Endo.Column.t()`s and `Endo.Table.t()`s.\n\nTakes an optional `Keyword.t()` of options:\n  - `timeout` which is an integer representing the number of milliseconds before which loading should be aborted.\n    This is only really a consideration for loading indexes across multiple tables and does not apply otherwise.\n    Defaults to `:timer.seconds(15)`.","ref":"Endo.Index.html#load/2","title":"Endo.Index.load/2","type":"function"},{"doc":"","ref":"Endo.Index.html#t:t/0","title":"Endo.Index.t/0","type":"type"},{"doc":"Utility module for surfacing adapter specific metadata","ref":"Endo.Metadata.html","title":"Endo.Metadata","type":"module"},{"doc":"Utility module for discovering Ecto Schema implementations for a given Endo Table","ref":"Endo.Schema.html","title":"Endo.Schema","type":"module"},{"doc":"Given a list of Endo Tables, tries to load their schemas in their corresponding OTP App.\n\nAll Endo Table structs contain metadata about which OTP app a given Repo belongs to. This information is used\nto load all Elixir modules that `use Ecto.Schema`.\n\nWith this list, we match up any schemas to tables that exist; though it is important to note that not all\nEndo Tables will necessarily have a corresponding Ecto Schema module defined for it.\n\nIn this case, the `schemas` key of an Endo Table will be an empty list.\n\nIt is also possible for multiple Ecto Schemas to exist for a single underlying database tables, thus, any discovered\nresults will be accumulated and returned as a list of modules per Endo Table.","ref":"Endo.Schema.html#load/1","title":"Endo.Schema.load/1","type":"function"},{"doc":"Table metadata returned by Endo.","ref":"Endo.Table.html","title":"Endo.Table","type":"module"},{"doc":"","ref":"Endo.Table.html#t:t/0","title":"Endo.Table.t/0","type":"type"},{"doc":"Exposes functions for some basic ETS operations.\nhttps://www.erlang.org/doc/man/ets.html\n# coveralls-ignore-start","ref":"Endo.Utils.ETS.html","title":"Endo.Utils.ETS","type":"module"},{"doc":"","ref":"Endo.Utils.ETS.html#from_list/1","title":"Endo.Utils.ETS.from_list/1","type":"function"},{"doc":"","ref":"Endo.Utils.ETS.html#get/3","title":"Endo.Utils.ETS.get/3","type":"function"},{"doc":"","ref":"Endo.Utils.ETS.html#has_key?/2","title":"Endo.Utils.ETS.has_key?/2","type":"function"},{"doc":"","ref":"Endo.Utils.ETS.html#new/1","title":"Endo.Utils.ETS.new/1","type":"function"},{"doc":"","ref":"Endo.Utils.ETS.html#put/3","title":"Endo.Utils.ETS.put/3","type":"function"},{"doc":"","ref":"Endo.Utils.ETS.html#to_list/1","title":"Endo.Utils.ETS.to_list/1","type":"function"},{"doc":"","ref":"Endo.Utils.ETS.html#t:type/0","title":"Endo.Utils.ETS.type/0","type":"type"}]}